@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2014  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -31,6 +31,7 @@
 #include "tree.h"
 #include "conversion.h"
 #include "socket.h"
+#include "sheduler.h"
 
 #ifndef STATIC		/* it's simpler to have dlcose() here */
 # ifdef HAVE_DLFCN_H
@@ -104,7 +105,7 @@ static FILE *lastdebuglog = NULL;
 
 /* lock for shutdown since it can be called any time */
 /* if LockIface is set then _Inum is implicitly locked for reading */
-pthread_mutex_t LockInum = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t LockInum = PTHREAD_MUTEX_INITIALIZER;
 
 static char PID_path[LONG_STRING];
 
@@ -210,6 +211,7 @@ void bot_shutdown (char *message, int e)
  */
 
 static pthread_mutex_t LockIface;
+static pthread_cond_t CondIface = PTHREAD_COND_INITIALIZER;
 
 ALLOCATABLE_TYPE (queue_i, _Q, next) /* alloc_queue_i(), free_queue_i() */
 
@@ -436,7 +438,7 @@ static void vsadd_request (ifi_t *to, iftype_t ift, const char *mask,
     if (Interface[i]->pq && Interface[i]->pq->request == cur)
     {
 #ifdef HAVE_ICONV
-      if (Interface[i]->a.conv)		/* if no conversion then skip it */
+      if (!(flag & F_RAW) && Interface[i]->a.conv) /* if no conversion then skip it */
       {
 	if (!conv)
 	{
@@ -598,7 +600,8 @@ static int _get_current (void)
   /* interface may be unused so lock semaphore */
   if (!Current->a.ift || (Current->a.ift & I_DIED))
     return 0;
-  time (&Time);			/* moved from sheduler to allow cycle */
+  Current->a.marked = FALSE;			/* drop mark right away */
+  Current->a.ift &= ~I_SLEEPING;
   if (!Current->a.IFRequest)
     out = REQ_OK;
   else if (curq)
@@ -625,6 +628,13 @@ static int _get_current (void)
   if (out == REQ_RELAYED && relay_request (curq->request))
     out = REQ_OK;
 
+  if (out == REQ_REJECTED)
+  {
+    DBG ("dispatcher:_get_current: interface %p suspended", &Current->a);
+    Current->a.marked = TRUE;			/* run it again next time */
+    Current->a.ift |= I_SLEEPING;		/* but with delay */
+  }
+
   if (out == REQ_OK)
     return delete_request (Current, curq);	/* else it was rejected */
 
@@ -755,6 +765,7 @@ static int _delete_iface (unsigned int r)
   register iftype_t rc;
 
   while (delete_request (curifi, curifi->head)); /* no queue for dead! */
+  _stop_timers (todel);
   if (!todel->IFSignal)			/* noone can be resumed from clone */
     ti = NULL;
   else for (ti = todel->prev; ti; ti = ti->prev)
@@ -899,10 +910,11 @@ static void iface_run (unsigned int i)
   }
   else if (!(Interface[i]->a.ift & I_LOCKED))
   {
-    register int gc;
+    register int gc = 0;
 
     stack_iface (&Interface[i]->a, 1);
-    gc = _get_current();		/* run with LockIface only */
+    if (Interface[i]->a.qsize > 0 || (Interface[i]->a.marked))
+      gc = _get_current();		/* run with LockIface only */
     while (gc > 0 && Interface[i]->a.qsize > 0)
       gc = _get_current();		/* try to empty queue at once */
     if (unstack_iface())
@@ -929,7 +941,7 @@ void Add_Request (iftype_t ift, const char *mask, flag_t fl, const char *text, .
     int savestate = O_GENERATECONF;
     register INTERFACE *li;
 
-    if (ift != -(iftype_t)1)	/* we assume only init will call it with -1 */
+    if (ift != (iftype_t)-1)	/* we assume only init will call it with -1 */
       O_GENERATECONF = FALSE;			/* don't make config now */
     inum = _Inum;				/* don't sent to created now! */
     if (Have_Wildcard (mask) < 0)
@@ -1041,11 +1053,27 @@ void dprint (int level, const char *text, ...)
   /* make pseudo "async-safe" connchain's debug when writing to socket */
   if (level <= O_DLEVEL && level < 9 && Interface && is_in_shutdown <= 0)
   {
-    pthread_mutex_lock (&LockIface);
-    vsadd_request (NULL, I_LOG, "*",
-		   F_DEBUG | (level < 1 ? F_ERROR : level == 1 ? F_WARN : 0),
-		   text, ap);
-    pthread_mutex_unlock (&LockIface);
+    int tries;
+
+    for (tries = 0; tries < 5; tries++)
+    {
+      /* since it might be called from a thread, safeguard on deadlock here */
+      if (pthread_mutex_trylock (&LockIface) == 0)
+      {
+	vsadd_request (NULL, I_LOG, "*",
+		       F_DEBUG | (level < 1 ? F_ERROR : level == 1 ? F_WARN : 0),
+		       text, ap);
+	pthread_mutex_unlock (&LockIface);
+	break;
+      }
+      else
+      {
+	struct timespec ts;
+	ts.tv_sec = 0;
+	ts.tv_nsec = 1000; /* sleep a microsecond */
+	nanosleep (&ts, NULL);
+      }
+    }
   }
   pthread_setcancelstate(cancelstate, NULL);
   va_end (ap);
@@ -1149,6 +1177,16 @@ int Rename_Iface (INTERFACE *iface, const char *newname)
   return 1;
 }
 
+/* locks on enter: none
+   should be called only when interface is guaranteed to stay still! */
+void Mark_Iface (INTERFACE *iface)
+{
+  if (unknown_iface (iface))
+    return;
+  iface->marked = TRUE;
+  pthread_cond_broadcast (&CondIface);		/* wake up dispatcher */
+}
+
 void Status_Interfaces (INTERFACE *iface)
 {
   register unsigned int i;
@@ -1281,15 +1319,36 @@ static int write_pid (int fd, pid_t pid)
   return 0;
 }
 
+static int sig_pipe[2]; /* pipe for signals delivery */
+
 static pthread_mutex_t SigLock = PTHREAD_MUTEX_INITIALIZER;
 
 static volatile sig_atomic_t _got_signal = 0;
 
+static void *sig_pipe_reader(void __attribute__((unused)) *data)
+{
+  sigset_t set;
+  int signo;
+
+  sigfillset(&set);
+  pthread_sigmask(SIG_BLOCK, &set, NULL);
+  FOREVER
+  {
+    if (read(sig_pipe[0], &signo, sizeof(int)) == sizeof(int))
+    {
+      pthread_mutex_lock (&SigLock);
+      _got_signal = signo;
+      pthread_cond_broadcast (&CondIface);	/* wake up dispatcher */
+      pthread_mutex_unlock (&SigLock);
+    }
+  }
+  /* never reached */
+  return NULL;
+}
+
 static void normal_handler (int signo)
 {
-  pthread_mutex_lock (&SigLock);
-  _got_signal = signo;
-  pthread_mutex_unlock (&SigLock);
+  while (write(sig_pipe[1], &signo, sizeof(int)) == 0);
 }
 
 static void errors_handler (int signo)
@@ -1343,12 +1402,13 @@ static void errors_handler (int signo)
 int dispatcher (INTERFACE *start_if)
 {
   struct sigaction act;
-  unsigned int i = 0, max = 0;
+  unsigned int i = 0;
   int limit = 0;
   int pidfd;
   pid_t pid;
   pthread_mutexattr_t attr;
   char *oldpid;
+  pthread_t sig_pipe_th;
 
   /* check if bot already runs */
   pidfd = set_pid_path();
@@ -1400,6 +1460,10 @@ int dispatcher (INTERFACE *start_if)
   if (freopen ("/dev/null", "w", stderr)) i=i;
   setsid();
   umask(0);
+  /* create a thread for signal pipe */
+  if (pipe (sig_pipe) != 0 ||
+      pthread_create(&sig_pipe_th, NULL, &sig_pipe_reader, NULL) != 0)
+    bot_shutdown (_("Failed to create signals thread."), 3);
   /* catch the signals */
   act.sa_handler = &normal_handler;
   sigemptyset (&act.sa_mask);
@@ -1480,6 +1544,10 @@ int dispatcher (INTERFACE *start_if)
   {
     if (_got_signal)
     {
+      char * argv[5];			/* required for restart */
+      char options[16];
+      int i;
+
       Set_Iface (NULL);			/* lock the dispatcher */
       /* Current is from last iface_run() so it should be valid even if dead */
       switch (_got_signal)
@@ -1494,13 +1562,26 @@ int dispatcher (INTERFACE *start_if)
 	  break;
 	case SIGINT:
 	  Add_Request (I_LOG, "*", F_BOOT, "Got SIGINT: restarting...");
-	  for (i = 0; i < _Inum; i++)		/* mark all listeners to die */
-	    if (Interface[i]->a.ift & I_CONNECT)
-	      Interface[i]->a.ift |= I_FINWAIT;	/* modules: unset flag later */
-	  ShutdownR = "Restart requested.";
-	  init();				/* restart */
-	  ShutdownR = NULL;
-	  break;
+	  /* stop everything */
+	  _bot_shutdown ("Restart requested.", 0);
+	  /* forge command line arguments */
+	  argv[0] = RunPath;
+	  argv[1] = options;
+	  options[0] = '-';
+	  if (O_DLEVEL > 5)
+	    O_DLEVEL = 5;
+	  for (i = 1; O_DLEVEL > 0; O_DLEVEL--)
+	    options[i++] = 'd';
+	  if (O_QUIET)
+	    options[i++] = 'q';
+	  if (O_DDLOG)
+	    options[i++] = 'D';
+	  options[i] = '\0';
+	  argv[2] = strrchr(Config, '/') + 1;
+	  argv[3] = NULL;
+	  /* re-exec itself */
+	  execv(argv[0], argv);
+	  /* should be never reached */
 	case SIGTERM:
 	default:
 	  bot_shutdown ("Got SIGTERM, shutdown...", 0);
@@ -1508,15 +1589,35 @@ int dispatcher (INTERFACE *start_if)
       _got_signal = 0;				/* reset state now */
       Unset_Iface();				/* continue if alive yet */
     }
-    pthread_mutex_lock (&LockIface);
-    max = _Inum;
-    pthread_mutex_unlock (&LockIface);
-    if (i >= max)
+    if (i >= _Inum)
     {
+      register int activity = 0; /* 0 - inactive, 1 - run, -1 - sleep&run */
+
+      pthread_mutex_lock (&LockIface);
       for (i = 0; i < _Inum; i++)
-	if (Interface[i]->a.qsize > 0)
-	  break;
-      PollSockets((i < _Inum) ? 1 : 0); /* sleep up to 200 ms */
+	if (Interface[i]->a.qsize > 0 || (Interface[i]->a.marked))
+	{					/* some interface needs care */
+	  if (activity <= 0 && (Interface[i]->a.ift & I_SLEEPING))
+	    activity = -1;
+	  else
+	    activity = 1;
+	}
+      pthread_mutex_unlock (&LockIface);
+      pthread_mutex_lock (&SigLock);		/* ensure _got_signal value */
+      if (_got_signal || activity > 0) ;	/* don't sleep */
+      else if (activity) {
+	struct timespec ts;
+
+	clock_gettime (CLOCK_REALTIME, &ts);
+	ts.tv_nsec += 500000000;		/* sleep for 0.5 second */
+	if (ts.tv_nsec >= 1000000000) {
+	  ts.tv_nsec -= 1000000000;
+	  ts.tv_sec++;
+	}
+	pthread_cond_timedwait (&CondIface, &SigLock, &ts);
+      } else
+	pthread_cond_wait (&CondIface, &SigLock); /* sleep for a while */
+      pthread_mutex_unlock (&SigLock);
       /* some cleanup stuff */
       i = 0;
     }
