@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2010  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -33,20 +33,31 @@ static char *EOL (const char *s)
   return strpbrk (s, "\r\n");
 }
 
+typedef struct HELPGR HELPGR;
+typedef struct HELPLANG HELPLANG;
+
 typedef struct HELP
 {
   char *key;			/* lower case */
   char *data;
-  void *helpgr;
+  HELPGR *helpgr;
   struct HELP *next;		/* in helpfile */
 } HELP;
 
-typedef struct HELPGR
+typedef struct HELPLGR
+{
+  struct HELPLGR *next;
+  HELPLANG *lang;
+  NODE *tree;
+} HELPLGR;
+
+struct HELPGR
 {
   char *key;			/* case-insensitive */
-  NODE *tree;
-  struct HELPGR *next;
-} HELPGR;
+  NODE *tree;			/* C lang */
+  HELPLGR *langs;
+  HELPGR *next;
+};
 
 typedef struct HELPFILE
 {
@@ -56,8 +67,15 @@ typedef struct HELPFILE
   struct HELPFILE *next;
 } HELPFILE;
 
+struct HELPLANG
+{
+  HELPLANG *next;
+  char *lang;			/* matchcase */
+  HELPFILE *files;
+};
+
 static HELPGR *Help = NULL;
-static HELPFILE *HFiles = NULL;
+static HELPLANG *HLangs = NULL;
 
 static HELPGR *_get_helpgr (char *name)
 {
@@ -80,41 +98,62 @@ static HELPGR *_get_helpgr (char *name)
   return gr;
 }
 
-int Add_Help (const char *name)
+static int Add_Help_L (const char *name, const char *lang)
 {
   char path[LONG_STRING];
   int fd = -1;
   off_t size;
+  HELPLANG **hl;
   HELPFILE **hf;
   HELP **ht;
+  NODE **tree;
+  HELPLGR **hlgr;
   char *data, *endc = NULL;
   char *key, *gr = NULL;
   register char *c;
   size_t s = 0;				/* compiler lies about uninitialized */
 
-  /* check if file.$locale exist */
-  if (*locale)
+  /* check if lang was already loaded */
+  for (hl = &HLangs; *hl; hl = &(*hl)->next)
+    if (safe_strcmp((*hl)->lang, lang) == 0)
+      break;
+  /* create new lang if it does not exist */
+  if (*hl == NULL)
   {
-    snprintf (path, sizeof(path), "%s/%s.%s", HELPDIR, name, locale);
+    *hl = safe_calloc(1, sizeof(HELPLANG));
+    (*hl)->lang = safe_strdup(lang);
+  }
+  /* check if name was already loaded for lang */
+  for (hf = &(*hl)->files; *hf; hf = &(*hf)->next)
+    if (safe_strcmp((*hf)->name, name) == 0)
+      break;
+  if (*hf)
+    /* already loaded, return success */
+    return 1;
+  /* check if file.$lang exists */
+  if (lang && *lang)
+  {
+    snprintf (path, sizeof(path), "%s/%s.%s", HELPDIR, name, lang);
     fd = open (path, O_RDONLY);
     /* check if file.$lang exists */
     if (fd < 0)
     {
-      snprintf (path, sizeof(path), "%s/%s.%.2s", HELPDIR, name, locale);
+      snprintf (path, sizeof(path), "%s/%s.%.2s", HELPDIR, name, lang);
       fd = open (path, O_RDONLY);
     }
   }
+  else
   /* check for default helpfile */
-  if (fd < 0)
   {
+    lang = "";
     snprintf (path, sizeof(path), "%s/%s", HELPDIR, name);
     fd = open (path, O_RDONLY);
-  }
-  /* last try: check if file.C exist */
-  if (fd < 0)
-  {
-    snprintf (path, sizeof(path), "%s/%s.C", HELPDIR, name);
-    fd = open (path, O_RDONLY);
+    /* last try: check if file.C exists */
+    if (fd < 0)
+    {
+      snprintf (path, sizeof(path), "%s/%s.C", HELPDIR, name);
+      fd = open (path, O_RDONLY);
+    }
   }
   if (fd < 0)
   /* may be it have to print error message? */
@@ -128,7 +167,6 @@ int Add_Help (const char *name)
     return 0;
   }
   lseek (fd, (off_t)0, SEEK_SET);
-  for (hf = &HFiles; *hf; hf = &(*hf)->next);	/* find the tail */
   *hf = safe_calloc (1, sizeof(HELPFILE));
   (*hf)->hfile = safe_malloc ((size_t)size + 1);
   if (read (fd, (*hf)->hfile, (size_t)size) != (ssize_t)size)
@@ -264,12 +302,26 @@ int Add_Help (const char *name)
 	else
 	  (*ht)->helpgr = _get_helpgr (gr);
 	(*ht)->key = key;
-	if (Insert_Key (&((HELPGR *)(*ht)->helpgr)->tree, key, *ht, 1))
-	  WARNING ("help: duplicate entry \"%s\" for set \"%s\" ignored", key,
-		   gr == key ? "" : gr);
+	if (*lang)
+	{
+	  for (hlgr = &(*ht)->helpgr->langs; *hlgr; hlgr = &(*hlgr)->next)
+	    if ((*hlgr)->lang == *hl)
+	      break;
+	  if (*hlgr == NULL)
+	  {
+	    *hlgr = safe_calloc(1, sizeof(HELPLGR));
+	    (*hlgr)->lang = *hl;
+	  }
+	  tree = &(*hlgr)->tree;
+	}
 	else
-	  dprint (2, "help: adding entry for \"%s\" to set \"%s\"", key,
-		  gr == key ? "" : gr);
+	  tree = &(*ht)->helpgr->tree;
+	if (Insert_Key (tree, key, *ht, 1))
+	  WARNING ("help: duplicate entry \"%s\" for set \"%s\" in lang \"%s\" ignored",
+		   key, gr == key ? "" : gr, lang);
+	else
+	  dprint (2, "help: adding entry for \"%s\" to set \"%s\" in lang \"%s\"",
+		  key, gr == key ? "" : gr, lang);
     }
     data = endc;
     if (data && *data == '\r')
@@ -282,50 +334,110 @@ int Add_Help (const char *name)
   return 1;
 }
 
-void Delete_Help (const char *name)
+int Add_Help (const char *name)
 {
-  HELPFILE *h, *hp = NULL;
-  HELP *t, *next = NULL;
+  int ret;
+  HELPLANG *hl;
+
+  if (HLangs == NULL)
+    return Add_Help_L (name, NULL);
+  else for (ret = 0, hl = HLangs; hl; hl = hl->next)
+    /* scan files for all already loaded languages */
+    if (Add_Help_L (name, hl->lang))
+      ret = 1;
+  return ret;
+}
+
+static void _delete_help_lang(HELPLANG *hl, const char *name)
+{
+  HELPFILE *h, **hp;
+  HELP *t, **tp;
+  HELPLGR *lgr;
+  NODE *tree;
 
   /* find the file */
-  for (h = HFiles; h; hp = h, h = h->next)
+  for (hp = &hl->files; (h = *hp); hp = &h->next)
   {
     if (!safe_strcmp (name, h->name))
       break;
   }
   if (!h)
     return;
-  if (hp)
-    hp->next = h->next;
-  else
-    HFiles = h->next;
+  *hp = h->next;
   /* delete keys from groups */
-  for (t = h->help; t; t = next)
+  for (tp = &h->help; (t = *tp); )
   {
-    next = t->next;
-    Delete_Key (((HELPGR *)t->helpgr)->tree, t->key, t);
-    dprint (2, "help: deleting entry for \"%s\" from set \"%s\"", t->key,
-	    NONULL(((HELPGR *)t->helpgr)->key));
+    tp = &t->next;
+    if (hl->lang) {
+      for (lgr = t->helpgr->langs; lgr; lgr = lgr->next)
+	if (lgr->lang == hl)
+	  break;
+      if (lgr == NULL)
+      {
+	/* this should never happen! */
+	ERROR("help: lang \"%s\" not found in set \"%s\"!", hl->lang,
+	      NONULL(t->helpgr->key));
+	continue;
+      }
+      tree = lgr->tree;
+    } else
+      tree = t->helpgr->tree;
+    Delete_Key (tree, t->key, t);
+    dprint (2, "help: deleting entry for \"%s\" from set \"%s\" in lang \"%s\"",
+	    t->key, NONULL(t->helpgr->key), NONULL(hl->lang));
     FREE (&t);
   }
-  Add_Request (I_LOG, "*", F_BOOT, "Unloaded helpfile %s", name);
+  Add_Request (I_LOG, "*", F_BOOT, "Unloaded helpfile %s (lang \"%s\")", name,
+	       NONULL(hl->lang));
   /* free memory */
   FREE (&h->hfile);
   FREE (&h->name);
   FREE (&h);
 }
 
-static int _no_such_help (INTERFACE *iface, int mode)
+void Delete_Help (const char *name)
 {
-  if (mode >= 0)
+  HELPLANG *hl;
+
+  /* process in each lang */
+  for (hl = HLangs; hl; hl = hl->next)
+    _delete_help_lang(hl, name);
+}
+
+/* tries to load language */
+static HELPLANG *_help_load_lang(const char *lang)
+{
+  HELPLANG *hl, **hlp;
+  HELPFILE *hf;
+
+  for (hlp = &HLangs; (hl = *hlp); hlp = &hl->next)
+    if (safe_strcmp(hl->lang, lang) == 0)
+      break;
+  if (hl == NULL && HLangs != NULL)
+  {
+    /* if not found then search and load $name.$lang for all added helpfiles */
+    *hlp = hl = safe_calloc(1, sizeof(HELPLANG));
+    hl->lang = safe_strdup(lang);
+    for (hf = HLangs->files; hf; hf = hf->next)
+      Add_Help_L(hf->name, lang);
+  }
+  return hl;
+}
+
+static int _no_such_help (INTERFACE *iface, const char *prefix, int each, int mode)
+{
+  if (mode < 0 || each < 0) ;
+  else if (each && prefix)
+    New_Request (iface, 0, "%s%s", prefix, _("No help on this."));
+  else
     New_Request (iface, 0, _("No help on this."));
   return 0;
 }
 
 #define HELP_LINE_SIZE 72
 
-static int _help_one_topic (char *text, INTERFACE *iface, char *prefix,
-			    const char *gr, const char *topic, int mode)
+static int _help_one_topic (char *text, INTERFACE *iface, const char *prefix,
+			    int each, const char *gr, const char *topic, int mode)
 {
   char buff[HUGE_STRING];
   char *c, *end;
@@ -339,7 +451,7 @@ static int _help_one_topic (char *text, INTERFACE *iface, char *prefix,
     if (text)
       text++;
     else
-      return _no_such_help (iface, mode);
+      return _no_such_help (iface, prefix, each, mode);
     if (*text == ':')
       mode--;
   }
@@ -375,13 +487,15 @@ static int _help_one_topic (char *text, INTERFACE *iface, char *prefix,
       *end++ = 0;
     /* may be, align all text with spaces here? :) */
     New_Request (iface, 0, "%s%s", prefix, c);
-    prefix = "  ";			/* indent next line :) */
+    if (!each)
+      prefix = "  ";			/* indent next line :) */
   }
   return 1;
 }
 
 static int _help_all_topics (HELPGR *what, INTERFACE *iface, userflag gf,
-			     userflag cf, struct bindtable_t *table, int mode)
+			     userflag cf, struct bindtable_t *table, int mode,
+			     const char *prefix, int each)
 {
   const char *key;
   LEAF *leaf = NULL;
@@ -397,7 +511,7 @@ static int _help_all_topics (HELPGR *what, INTERFACE *iface, userflag gf,
       if (strcmp(key, b->key))		/* it's matched but just similar */
 	continue;
       /* is this first? */
-      if (!r)
+      if (!r && each == 0)
       {
 	/* mode dependent message? */
 	New_Request (iface, 0, _("Available topics for \"help%s%s\":"),
@@ -410,12 +524,12 @@ static int _help_all_topics (HELPGR *what, INTERFACE *iface, userflag gf,
       {
 	if (s)
 	{
-	  New_Request (iface, 0, "%s", buf);
+	  New_Request (iface, 0, "%s%s", (each && prefix) ? prefix : "", buf);
 	  s = 0;
 	}
 	if (ns >= HELP_LINE_SIZE)
 	{
-	  New_Request (iface, 0, "%s", key);
+	  New_Request (iface, 0, "%s%s", (each && prefix) ? prefix : "", key);
 	  continue;
 	}
       }
@@ -426,36 +540,47 @@ static int _help_all_topics (HELPGR *what, INTERFACE *iface, userflag gf,
     }
   }
   if (s)
-    New_Request (iface, 0, "%s", buf);
+    New_Request (iface, 0, "%s%s", (each && prefix) ? prefix : "", buf);
   return r;
 }
 
-int Get_Help (const char *fst, const char *sec, INTERFACE *iface, userflag gf,
-	      userflag cf, struct bindtable_t *table, char *prefix, int mode)
+#ifndef ENABLE_NLS
+static
+#endif
+int Get_Help_L (const char *fst, const char *sec, INTERFACE *iface, userflag gf,
+		userflag cf, struct bindtable_t *table, const char *prefix,
+		int each, int mode, const char *lang)
 {
   const char *topic = NONULL(sec);	/* default topic=fst{sec} */
   HELPGR *h = Help;
-  HELP *t;
+  HELP *t = NULL;
+  HELPLANG *hl;
+  HELPLGR *hlgr;
 
   dprint (5, "help:Get_Help: call \"%s %s\"", NONULL(fst), NONULL(sec));
   if (!h)
-    return _no_such_help (iface, mode);		/* no help loaded */
+    return _no_such_help (iface, prefix, each, mode);	/* no help loaded */
   else if (!table)
   {
     if (!fst)
-      return _no_such_help (iface, mode);	/* NULL/smth : no table */
+      return _no_such_help (iface, prefix, each, mode);	/* NULL/smth : no table */
   }
   else if (!fst && *topic)					/* NULL smth */
   {
     if (strcmp (topic, "*") &&					/* NULL !"*" */
 	Check_Bindtable (table, sec, gf, cf, NULL) == NULL)
-      return _no_such_help (iface, mode);	/* NULL/name : not found */
+      return _no_such_help (iface, prefix, each, mode);	/* NULL/name : not found */
     fst = Bindtable_Name (table);		/* -> table/name */
   }
   else if (!fst || !*fst || !strcmp (fst, "*"))			/* "*" NULL */
-    return _help_all_topics (Help, iface, gf, cf, table, mode);
+    return _help_all_topics (Help, iface, gf, cf, table, mode, prefix, each);
+  else if (*fst == '=')
+  {
+    if (Check_Bindtable (table, topic, gf, cf, NULL) == NULL)
+      return _no_such_help (iface, prefix, each, mode);
+  }
   else if (Check_Bindtable (table, fst, gf, cf, NULL) == NULL)	/* smth "*" */
-    return _no_such_help (iface, mode);		/* no fst in table */
+    return _no_such_help (iface, prefix, each, mode);	/* no fst in table */
   /* check for group first - if is second parameter! */
   if (!*topic)					/* if fst/NULL */
     topic = fst;			/* then topic=fst{fst} */
@@ -467,22 +592,38 @@ int Get_Help (const char *fst, const char *sec, INTERFACE *iface, userflag gf,
   /* no such group? */
   if (!h)
   {
-    WARNING ("help: set \"%s\" not found", fst);
+    dprint (4, "help: set \"%s\" not found", fst);
     if (sec && *sec)
-      return _no_such_help (iface, mode);	/* BAD/smth */
+      return _no_such_help (iface, prefix, each, mode);	/* BAD/smth */
     /* find common help */
     h = Help;					/* if smth/NULL */
   }
   if (table && !strcmp (topic, "*"))		/* if topic is ...{*} */
-    return _help_all_topics (h, iface, gf, cf, table, mode);
-  t = Find_Key (h->tree, topic);
+    return _help_all_topics (h, iface, gf, cf, table, mode, prefix, each);
+  if (lang && *lang)
+  {
+    hl = _help_load_lang(lang);
+    for (hlgr = h->langs; hlgr; hlgr = hlgr->next)
+      if (hlgr->lang == hl)
+	break;
+    if (hlgr)
+      t = Find_Key (hlgr->tree, topic);
+  }
+  if (t == NULL)
+    t = Find_Key (h->tree, topic);
   if (!t)
   {
-    WARNING ("help: topic \"%s\" not found", topic);
+    dprint (4, "help: topic \"%s\" not found", topic);
     if (table && h != Help && !sec)		/* table/NULL */
-      return _help_all_topics (h, iface, gf, cf, table, mode);
-    return _no_such_help (iface, mode);
+      return _help_all_topics (h, iface, gf, cf, table, mode, prefix, each);
+    return _no_such_help (iface, prefix, each, mode);
   }
   dprint (4, "help: found entry for \"%s\" in set \"%s\"", topic, NONULL(h->key));
-  return _help_one_topic (t->data, iface, NONULL(prefix), h->key, topic, mode);
+  return _help_one_topic (t->data, iface, NONULL(prefix), each, h->key, topic, mode);
+}
+
+int Get_Help (const char *fst, const char *sec, INTERFACE *iface, userflag gf,
+	      userflag cf, struct bindtable_t *table, const char *prefix, int mode)
+{
+  return Get_Help_L (fst, sec, iface, gf, cf, table, prefix, 0, mode, locale);
 }
