@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -406,7 +406,7 @@ struct binding_t *Check_Bindtable (struct bindtable_t *bt, const char *str,
       l = Find_Leaf (bt->list.tree, "", 0);
     return (l == NULL) ? NULL : l->s.data;
   }
-  cf = (scf & ~U_EQUAL);		/* drop the flag to matching */
+  cf = (scf & ~(U_EQUAL | U_NEGATE));	/* drop the flag to matching */
   if (bt->type == B_MASK || bt->type == B_UNIQMASK)
     cc = 0;
   else if (bt->type == B_MATCHCASE)
@@ -450,11 +450,15 @@ struct binding_t *Check_Bindtable (struct bindtable_t *bt, const char *str,
     }
     else if (b)
     {
-      if (b->gl_uf & U_NEGATE)				/* -a */
+      if (gf & U_NEGATE)				/* !a */
+	tgf = ~b->gl_uf;
+      else if (b->gl_uf & U_NEGATE)			/* -a */
 	tgf = (b->gl_uf & ~gf);
       else						/* a */
 	tgf = (b->gl_uf & gf);
-      if (b->ch_uf & U_NEGATE)				/* -b */
+      if (scf & U_NEGATE)				/* !b */
+	tcf = ~b->ch_uf;
+      else if (b->ch_uf & U_NEGATE)			/* -b */
 	tcf = (b->ch_uf & ~cf);
       else						/* b */
 	tcf = (b->ch_uf & cf);
@@ -480,11 +484,15 @@ struct binding_t *Check_Bindtable (struct bindtable_t *bt, const char *str,
     }
     else while (b)
     {
-      if (b->gl_uf & U_NEGATE)				/* -a */
+      if (gf & U_NEGATE)				/* !a */
+	tgf = ~b->gl_uf;
+      else if (b->gl_uf & U_NEGATE)			/* -a */
 	tgf = (b->gl_uf & ~gf);
       else						/* a */
 	tgf = (b->gl_uf & gf);
-      if (b->ch_uf & U_NEGATE)				/* -b */
+      if (scf & U_NEGATE)				/* !b */
+	tcf = ~b->ch_uf;
+      else if (b->ch_uf & U_NEGATE)			/* -b */
 	tcf = (b->ch_uf & ~cf);
       else						/* b */
 	tcf = (b->ch_uf & cf);
@@ -518,11 +526,15 @@ struct binding_t *Check_Bindtable (struct bindtable_t *bt, const char *str,
 	i = 1;				/* support for U_COMPL type */
 	break;				/* found! */
       }
-      if (b->gl_uf & U_NEGATE)				/* -a */
+      if (gf & U_NEGATE)				/* !a */
+	tgf = ~b->gl_uf;
+      else if (b->gl_uf & U_NEGATE)			/* -a */
 	tgf = (b->gl_uf & ~gf);
       else						/* a */
 	tgf = (b->gl_uf & gf);
-      if (b->ch_uf & U_NEGATE)				/* -b */
+      if (scf & U_NEGATE)				/* !b */
+	tcf = ~b->ch_uf;
+      else if (b->ch_uf & U_NEGATE)			/* -b */
 	tcf = (b->ch_uf & ~cf);
       else						/* b */
 	tcf = (b->ch_uf & cf);
@@ -578,7 +590,7 @@ struct binding_t *Check_Bindtable (struct bindtable_t *bt, const char *str,
 
 const char *Bindtable_Name (struct bindtable_t *bt)
 {
-  if (!bt)
+  if (bt == NULL)
     return NULL;
   return bt->name;
 }
@@ -1923,10 +1935,12 @@ static int dc_chelp(struct peer_t *dcc, char *args)
   if (args == NULL || *args == '\0')	/* should have a parameter */
     return 0;
   /* try if usage found first */
-  if (Get_Help ("function", args, dcc->iface, dcc->uf, 0, NULL, _("Usage: "), 0))
+  if (Get_Help_L ("function", args, dcc->iface, dcc->uf, 0, NULL, _("Usage: "), 0, 0,
+		  dcc->lang[0] ? dcc->lang : locale))
     return 0;
   /* full help */
-  Get_Help ("function", args, dcc->iface, dcc->uf, 0, NULL, NULL, 2);
+  Get_Help_L ("function", args, dcc->iface, dcc->uf, 0, NULL, NULL, 0, 2,
+	      dcc->lang[0] ? dcc->lang : locale);
   return 1;
 }
 
@@ -1936,6 +1950,7 @@ static int dc_chelp(struct peer_t *dcc, char *args)
 
 static struct bindtable_t *BT_IsOn = NULL;
 static struct bindtable_t *BT_Inspect = NULL;
+static struct bindtable_t *BT_Update = NULL;
 
 int Lname_IsOn (const char *net, const char *pub, const char *lname,
 		const char **name)
@@ -1994,6 +2009,36 @@ modeflag Inspect_Client (const char *net, const char *pub, const char *name,
   return f (net, pub, name, lname, host, idle, cnt);
 }
 
+int Update_Public (const char *net, const char *pub, modeflag change,
+		   const char *tgt, const char *topic, const char *by, time_t at)
+{
+  struct binding_t *bind;
+  const char *nt;
+  struct clrec_t *netw;
+#define static register
+  BINDING_TYPE_update_public ((*f));
+#undef static
+
+  if (!net)
+    return 0;
+  netw = Lock_Clientrecord (net);
+  if (netw)
+  {
+    if ((Get_Flags (netw, NULL) & U_SPECIAL) &&
+	(nt = Get_Field (netw, ".logout", NULL))) /* get network type */
+      bind = Check_Bindtable (BT_Update, nt, U_ALL, U_ANYCH, NULL);
+    else
+      bind = NULL;
+    Unlock_Clientrecord (netw);
+  }
+  else
+    bind = NULL;
+  if (!bind || bind->name)
+    return 0;		/* no such network/service */
+  f = bind->func;
+  return f (net, pub, change, tgt, topic, by, at);
+}
+
 /* ----------------------------------------------------------------------------
  * main init()
  */
@@ -2089,6 +2134,7 @@ INTERFACE *init (void)
   BT_Script = Add_Bindtable ("script", B_UNIQMASK);
   BT_IsOn = Add_Bindtable ("ison", B_UNIQ);
   BT_Inspect = Add_Bindtable ("inspect-client", B_UNIQ);
+  BT_Update = Add_Bindtable ("update-public", B_UNIQ);
 
   _add_fn ("set", &cfg_set, NULL);		/* add operators */
   _add_fn ("script", &cfg_script, NULL);	/* but don't ask */
@@ -2159,6 +2205,7 @@ INTERFACE *init (void)
     ConfigFileIface->ift = I_TEMP;		/* force flags */
     Set_Iface (ConfigFileIface);		/* it's current again */
     _unsharp_var("charset");
+    _unsharp_var("locale");
     /* it's time to write to config so signal to register again */
     Send_Signal (-1, "*", S_REG);
     Set_Iface (Init);				/* flush everything now */
