@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -219,19 +219,39 @@ static struct clrec_t *_findthebest (const char *mask, struct clrec_t *prefer)
 {
   struct clrec_t *u, *user = NULL;
   user_hr *hr;
+  char *at_excl;
   int n, p = 0, matched = 0;
   lid_t lid;
   char lcmask[HOSTMASKLEN+1];
+  char hostmask[HOSTMASKLEN+1];
+  register const char *c;
 
-  unistrlower (lcmask, mask, sizeof(lcmask));  
+  unistrlower (lcmask, mask, sizeof(lcmask));
+  at_excl = strchr(lcmask, '!');
   rw_rdlock (&HLock);
   lid = LID_MIN;
   do {
     if ((u = UList[lid - LID_MIN]) && !(u->flag & (U_SPECIAL|U_ALIAS)))
-    /* pseudo-users hosts are not masks */
       for (hr = u->host; hr; hr = hr->next)
       {
-	n = match (hr->hostmask, lcmask);
+	/* skip xxx! part in hostmask if input does not contain one yet */
+	if (!at_excl && (c = strchr(hr->hostmask, '!')))
+	  c++;
+	else
+	  c = hr->hostmask;
+	for (n = 0; *c && n < (int)sizeof(hostmask)-1; )
+	{
+	  /* special support for masks with password and port */
+	  if (*c == ':')
+	    while (*c && *c != '@') c++;
+	  if (*c == '@') {
+	    while (*c && *c != '/' && n < (int)sizeof(hostmask)-1) hostmask[n++] = *c++;
+	    break;
+	  } else
+	    hostmask[n++] = *c++;
+	}
+	hostmask[n] = '\0';
+	n = match (hostmask, lcmask);
 	/* find max */
 	if (n > matched)
 	{
@@ -276,15 +296,17 @@ static struct clrec_t *_findbyhost (const char *mask)
   return u;
 }
 
-static int _check_subpattern(char **ptr, char end1, char end2)
+static int _check_subpattern(const char **ptr, char end1, char end2)
 {
-  char *c;
+  const char *c;
   int r = 0, r1, r2;
 
   for (c = *ptr; *c && *c != end1 && *c != end2; c++)
   {
     if (*c == '*' || *c == '?') /* wildcard */
       continue;
+    if (*c == ':') /* delimiter of password or IPv6 parts */
+      continue;
     if (*c == '[') /* range */
     {
       if (*c == '-')
@@ -322,21 +344,23 @@ static int _add_usermask (struct clrec_t *user, const char *mask)
   user_hr **h;
   int r;
   char lcmask[HOSTMASKLEN+1];
-  char *c = lcmask;
+  const char *c = mask;
+  const char *m = mask;
 
-  if (user->uid > ID_ANY && strchr(user->lname, '.'))
-    /* it's server name rather than user name */
-    strfcpy (lcmask, mask, sizeof(lcmask));
-  else
-    unistrlower (lcmask, mask, sizeof(lcmask));
   /* test the pattern to contain not just wildcards */
   r = _check_subpattern(&c, '!', '@');
   if (*c == '!')
     c++, r += _check_subpattern(&c, '@', 0);
   if (*c == '@')
-    c++, r += _check_subpattern(&c, 0, 0);
+    m = ++c, r += _check_subpattern(&c, 0, 0);
   if (r < 1)			/* at least a single non-wildcard char required */
     return 0;
+  if (m - mask > IDENTLEN + PASSWDMAX) /* too big i:p part of the mask */
+    return 0;
+  if (m > mask)
+    /* keep ident + password intact */
+    memmove (lcmask, mask, m - mask);
+  unistrlower (lcmask + (m - mask), m, sizeof(lcmask) - (m - mask));
   /* check for aliases */
   if (user->flag & U_ALIAS)	/* no need lock since threads has R/O access */
     user = user->u.owner;
@@ -948,7 +972,7 @@ int Get_Clientlist (INTERFACE *iface, userflag uf, const char *fn,
   canbenonamed = uf & U_NONAMED;
   fnisservice = safe_strchr ((char *)fn, '@');	/* isn't NULL if service name */
   gf = uf & U_GLOBALS;
-  unistrlower (lcmask, mask, sizeof(lcmask));  
+  unistrlower (lcmask, mask, sizeof(lcmask));
   n = 0;
   len = 0;
   lid = LID_MIN;
@@ -959,7 +983,7 @@ int Get_Clientlist (INTERFACE *iface, userflag uf, const char *fn,
       if (canbenonamed ||		/* if it's check for ban/invite/etc. */
 	  fn == NULL)			/* NULL means check lname and host */
       {
-	if (match (lcmask, u->lclname) >= 0)	/* Lname matched to it */
+	if (u->lname && match (lcmask, u->lclname) >= 0) /* Lname matches */
 	  n += _add_to_list (iface, buf, &len, u->lname);
 	else for (h = u->host; h; h = h->next)
 	  if (match (lcmask, h->hostmask) > 0)	/* hostmask matched to it */
@@ -2858,8 +2882,7 @@ static int dc__phost (struct peer_t *dcc, char *args)
   if (!user)				/* no user */
     return 0;
   args = NextWord (args);
-  if (strlen (args) < 5 || match ("*.*", args) < 0 ||
-      _add_usermask (user, args) == 0) /* validation of hostmask */
+  if (_add_usermask (user, args) == 0) /* validation of hostmask */
   {
     New_Request (dcc->iface, 0, "Invalid hostmask pattern: %s", args);
     return 0;
