@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2011  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -18,6 +18,9 @@
  * This file is part of FoxEye's source: common declarations.
  */
 
+#ifndef _PROTOS_H
+#define _PROTOS_H 1
+
 #define uchar unsigned char
 
 /* ----------------------------------------------------------------------------
@@ -30,6 +33,7 @@ INTERFACE *Find_Iface (iftype_t, const char *) __attribute__((warn_unused_result
 INTERFACE *Set_Iface (INTERFACE *);
 int Unset_Iface (void);
 int Rename_Iface (INTERFACE *, const char *);
+void Mark_Iface (INTERFACE *);
 void Add_Request (iftype_t, const char *, flag_t, const char *, ...)
 	__attribute__((format(printf, 4, 5)));
 void New_Request (INTERFACE *, flag_t, const char *, ...)
@@ -54,10 +58,18 @@ int Lname_IsOn (const char *, const char *, const char *, const char **)
 modeflag Inspect_Client (const char *, const char *, const char *,
 			 const char **, const char **, time_t *, short *)
 	__attribute__((warn_unused_result));
+int Update_Public (const char *, const char *, modeflag, const char *,
+		   const char *, const char *, time_t);
 int Add_Help (const char *) __attribute__((nonnull(1))); /* help.c */
 void Delete_Help (const char *);
 int Get_Help (const char *, const char *, INTERFACE *, userflag, userflag,
-		struct bindtable_t *, char *, int);
+		struct bindtable_t *, const char *, int);
+#ifdef ENABLE_NLS
+int Get_Help_L (const char *, const char *, INTERFACE *, userflag, userflag,
+		struct bindtable_t *, const char *, int, int, const char *);
+#else
+#define Get_Help_L(a,b,c,d,e,f,g,h,i,j) Get_Help(a,b,c,d,e,f,g,i)
+#endif
 int Add_Clientrecord (const char *, const uchar *, userflag) /* users.c */
 	__attribute__((warn_unused_result));
 int Add_Alias (const char *, const char *)
@@ -294,3 +306,4 @@ static inline int clock_gettime(int clk_id, struct timespec* t)
     return 0;
 }
 #endif
+#endif /* _PROTOS_H */
