@@ -449,7 +449,6 @@ static void isend_handler (char *lname, char *ident, const char *host, void *dat
   scd = safe_calloc(1, sizeof(struct send_cleanup));
   scd->socket = dcc->socket;
   scd->buff = buff = safe_malloc (MAXBLOCKSIZE);
-  pthread_cleanup_push(&_send_handler_cleanup, scd);
   Set_Iface (dcc->l.iface);
   if (host)				/* if it's from passive then it's NULL */
   {
@@ -457,6 +456,7 @@ static void isend_handler (char *lname, char *ident, const char *host, void *dat
     snprintf (&dcc->uh[sr], sizeof(dcc->uh) - sr, "!%s@%s", ident ? ident : "*",
 	      host ? host : "*");
   }					/* dcc->uh is nick!user@host now */
+  pthread_cleanup_push(&_send_handler_cleanup, scd);
   dcc->ptr = 0;
   dcc->rate = 0;
   if (dcc->startptr == 1)		/* it was '-flush' flag */
@@ -1177,8 +1177,8 @@ static void *_dcc_stage_1 (void *input_data)
     dcc->wait_accept = TRUE;		/* let thread know about resume */
     dcc->startptr = 0;
     dcc->state = P_IDLE;		/* interface is still alive! */
-    dcc->tid = NewTimer (I_CONNECT, dcc->l.iface->name, S_TIMEOUT,
-			 ircdcc_resume_timeout, 0, 0, 0); /* timeout for ACCEPT */
+    dcc->tid = Add_Timer (dcc->l.iface, S_TIMEOUT, ircdcc_resume_timeout);
+					/* timeout for ACCEPT */
     Unset_Iface();
     pthread_exit(NULL);
   }
@@ -1856,7 +1856,7 @@ static int ssirc_send (struct peer_t *peer, INTERFACE *w, char *args)
 		 net, ip_local, dcc->size, _ircdcc_dccid);
     snprintf (target, sizeof(target), "irc-ctcp#%u", _ircdcc_dccid++);
     dcc->l.iface = Add_Iface (I_TEMP, target, &_isend_sig_w, NULL, dcc);
-    dcc->tid = NewTimer (I_TEMP, target, S_TIMEOUT, ircdcc_conn_timeout, 0, 0, 0);
+    dcc->tid = Add_Timer (dcc->l.iface, S_TIMEOUT, ircdcc_conn_timeout);
     /* done... we should wait for responce now so we can connect there */
   } else {
     struct dcc_listener_data *dld;
