@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -24,12 +24,15 @@
 #include <list.h>
 #include <init.h>
 #include <conversion.h>
+#include <socket.h>
 
 #include "ircd.h"
 #include "numerics.h"
 
 extern bool _ircd_public_topic; /* in ircd.c */
 
+static struct bindtable_t *BTIrcdClientCmd; /* copy of pointer in ircd.c */
+
 /* ---------------------------------------------------------------------------
  * Common internal functions.
  */
@@ -164,7 +167,7 @@ static int ircd_oper_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
     cl->umode |= A_OP;			/* global oper */
     New_Request (peer->iface, 0, ":%s MODE %s +o", peer->dname, peer->dname);
     ircd_sendto_servers_new ((IRCD *)srv->data, NULL, ":%s IMODE %d %s +o",
-			     peer->dname, ircd_new_id(), peer->dname);
+			     peer->dname, ircd_new_id(NULL), peer->dname);
 #ifdef USE_SERVICES
     ircd_sendto_services_mark_nick ((IRCD *)srv->data, SERVICE_WANT_OPER);
 #endif
@@ -229,7 +232,8 @@ static int ircd_squit_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick
 			 const char *user, const char *host, const char *vhost,
 			 modeflag eum, int argc, const char **argv)
 { /* args: <server> <comment> */
-  CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
+  struct peer_priv *pp = peer->iface->data; /* it's really peer */
+  CLIENT *cl = pp->link->cl;
   CLIENT *tgt;
 
   if (argc < 2)
@@ -245,15 +249,16 @@ static int ircd_squit_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick
       ircd_do_unumeric (cl, ERR_NOSUCHSERVER, cl, 0, argv[0]);
     return ircd_do_unumeric (cl, ERR_NOPRIVILEGES, cl, 0, NULL);
   }
-  if (!tgt || CLIENT_IS_ME(tgt) || !CLIENT_IS_SERVER(tgt))
+  if (!tgt || tgt->hold_upto || CLIENT_IS_ME(tgt) || !CLIENT_IS_SERVER(tgt))
     return ircd_do_unumeric (cl, ERR_NOSUCHSERVER, cl, 0, argv[0]);
   /* we doing squit only for shortest way despite of possible multiconnect! */
   if (CLIENT_IS_LOCAL(tgt)) {		/* squit if it's local link */
     ircd_sendto_wallops((IRCD *)srv->data, NULL, me, "SQUIT %s from %s: %s",
 			argv[0], cl->nick, argv[1]);
     ircd_do_squit (tgt->via->link, NULL, argv[1]); /* do job */
-  } else				/* or else forward it to it's links */
-    ircd_sendto_remote (tgt, ":%s SQUIT %s :%s", peer->dname, argv[0], argv[1]);
+  } else {				/* or else forward it to its link */
+    ircd_sendto_one (tgt, ":%s SQUIT %s :%s", peer->dname, argv[0], argv[1]);
+  }
   return 1;
 }
 
@@ -403,7 +408,7 @@ static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick
 #endif
     ircd_sendto_servers_mask_new ((IRCD *)srv->data, NULL, cmask,
 				  ":%s ITOPIC %d %s :%s", peer->dname,
-				  ircd_new_id(), ch->name, ch->topic);
+				  ircd_new_id(NULL), ch->name, ch->topic);
     ircd_sendto_servers_mask_old ((IRCD *)srv->data, NULL, cmask,
 				  ":%s TOPIC %s :%s", peer->dname, ch->name,
 				  ch->topic);
@@ -415,7 +420,7 @@ static int ircd_topic_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick
   ircd_sendto_servers_old ((IRCD *)srv->data, NULL, ":%s TOPIC %s :%s",
 			   peer->dname, ch->name, ch->topic);
   ircd_sendto_servers_new ((IRCD *)srv->data, NULL, ":%s ITOPIC %d %s :%s",
-			   peer->dname, ircd_new_id(), ch->name, ch->topic);
+			   peer->dname, ircd_new_id(NULL), ch->name, ch->topic);
   return 1;
 }
 
@@ -446,7 +451,7 @@ static int ircd_invite_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnic
   if (CLIENT_IS_REMOTE(tgt)) {
     if ((memb != NOSUCHCHANNEL) && (memb->chan->mode & A_INVISIBLE))
       return (1); /* TODO: send some numeric to client? */
-    ircd_sendto_one (tgt, ":%s INVITE %s %s", peer->dname, argv[0], argv[1]);
+    ircd_sendto_remote (tgt, cl, NULL, ":%s INVITE %s %s", peer->dname, argv[0], argv[1]);
   } else
     ircd_sendto_one (tgt, ":%s!%s@%s INVITE %s %s", peer->dname, user, vhost,
 		     argv[0], argv[1]);
@@ -556,11 +561,11 @@ static int ircd_servlist_cb(INTERFACE *srv, struct peer_t *peer, const char *lcn
     if (!tgt->hold_upto && CLIENT_IS_SERVICE (tgt) &&
 	simple_match (mask, tgt->lcnick) >= 0)
     {
-      snprintf (buf, sizeof(buf), "%s %s %s %hu :%s",
+      snprintf (buf, sizeof(buf), "%s %s %s %d :%s",
 #ifdef USE_SERVICES
       !CLIENT_IS_REMOTE (tgt) ? me->nick :
 #endif
-		tgt->cs->nick, mask, tgt->away, tgt->hops - 1, tgt->fname);
+		tgt->cs->nick, mask, tgt->away, (int)tgt->hops - 1, tgt->fname);
       ircd_do_unumeric (cl, RPL_SERVLIST, tgt, 0, buf);
     }
   }
@@ -575,23 +580,25 @@ static inline void _ircd_who_reply (CLIENT *rq, CLIENT *srv, CLIENT *tgt,
 
   if (m)
   {
-    char ch[2];
+    char ch[8];
 
     ch[1] = 0;
-    if (m->mode & A_ADMIN)
+    if (m->mode & A_ADMIN) {
       ch[0] = '@';
-    else
-      ch[0] = ircd_mode2whochar (m->mode);
-    snprintf (buf, sizeof(buf), "%s %s %s %s %s %c%s%s :%hu %s", m->chan->name,
+      ch[1] = 0;
+    } else
+      ircd_mode2whochar (m->mode, ch, sizeof(ch));
+    snprintf (buf, sizeof(buf), "%s %s %s %s %s %c%s%s :%d %s", m->chan->name,
 	      tgt->user, tgt->vhost, srv->lcnick, tgt->nick,
 	      (tgt->umode & A_AWAY) ? 'G' : 'H',
-	      (tgt->umode & (A_OP | A_HALFOP)) ? "*" : "", ch, tgt->hops - 1,
+	      (tgt->umode & (A_OP | A_HALFOP)) ? "*" : "", ch, (int)tgt->hops - 1,
 	      tgt->fname);
   }
   else
-    snprintf (buf, sizeof(buf), "* %s %s %s %s %c%s :%hu %s", tgt->user,
+    snprintf (buf, sizeof(buf), "* %s %s %s %s %c%s :%d %s", tgt->user,
 	      tgt->vhost, srv->lcnick, tgt->nick, (tgt->umode & A_AWAY) ? 'G' : 'H',
-	      (tgt->umode & (A_OP | A_HALFOP)) ? "*" : "", tgt->hops - 1, tgt->fname);
+	      (tgt->umode & (A_OP | A_HALFOP)) ? "*" : "", (int)tgt->hops - 1,
+	      tgt->fname);
   ircd_do_unumeric (rq, RPL_WHOREPLY, rq, 0, buf);
 }
 
@@ -740,7 +747,7 @@ static int ircd_away_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
     cl->away[0] = '\0';
     cl->umode &= ~A_AWAY;
     ircd_sendto_servers_new((IRCD *)srv->data, NULL, ":%s IMODE %d %s :-a",
-			    peer->dname, ircd_new_id(), peer->dname);
+			    peer->dname, ircd_new_id(NULL), peer->dname);
     ircd_sendto_servers_old((IRCD *)srv->data, NULL, ":%s MODE %s :-a",
 			    peer->dname, peer->dname);
 #ifdef USE_SERVICES
@@ -755,7 +762,7 @@ static int ircd_away_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
   strfcpy(cl->away, argv[0], len + 1); /* unistrcut includes '\0' */
   cl->umode |= A_AWAY;
   ircd_sendto_servers_new((IRCD *)srv->data, NULL, ":%s IMODE %d %s :+a",
-			  peer->dname, ircd_new_id(), peer->dname);
+			  peer->dname, ircd_new_id(NULL), peer->dname);
   ircd_sendto_servers_old((IRCD *)srv->data, NULL, ":%s MODE %s :+a",
 			  peer->dname, peer->dname);
 #ifdef USE_SERVICES
@@ -860,7 +867,10 @@ static int ircd_userhost_cb(INTERFACE *srv, struct peer_t *peer, const char *lcn
 	buf[s++] = '+';
       s += strfcpy(&buf[s], tgt->user, sizeof(buf) - s);
       buf[s++] = '@';
-      s += strfcpy(&buf[s], tgt->vhost, sizeof(buf) - s);
+      if (tgt == cl)
+	s += strfcpy(&buf[s], SocketIP(peer->socket), sizeof(buf) - s);
+      else
+	s += strfcpy(&buf[s], tgt->vhost, sizeof(buf) - s);
     }
   }
   if (s)
@@ -912,11 +922,28 @@ static int ircd_charset_cb(INTERFACE *srv, struct peer_t *peer, const char *lcni
 { /* args: [<charset>] */
   CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl;
   struct conversion_t *conv;
+  char u_nick[MB_LEN_MAX*NICKLEN+1];
+  char n_nick[MB_LEN_MAX*NICKLEN+1];
+  char *u_nick_ptr = u_nick;
+  char *n_nick_ptr = n_nick;
+  size_t s;
 
   if (argc > 0) {
     conv = Get_Conversion(argv[0]);
     if (conv == NULL && strcasecmp(Conversion_Charset(conv), argv[0]))
       return ircd_do_unumeric(cl, ERR_NOCODEPAGE, cl, 0, argv[0]);
+    /* prepare :<old charset nick> NICK <new charset nick> and send with F_RAW */
+    s = strlen(peer->dname);
+    s = Undo_Conversion(peer->iface->conv, &u_nick_ptr, sizeof(u_nick) - 1,
+			peer->dname, &s);
+    if (u_nick_ptr == u_nick)
+      u_nick[s] = 0;
+    s = strlen(peer->dname);
+    s = Undo_Conversion(conv, &n_nick_ptr, sizeof(n_nick) - 1, peer->dname, &s);
+    if (n_nick_ptr == n_nick)
+      n_nick[s] = 0;
+    if (strcmp(u_nick_ptr, n_nick_ptr) != 0)
+      New_Request(peer->iface, F_RAW, ":%s NICK %s", u_nick_ptr, n_nick_ptr);
     Free_Conversion(peer->iface->conv);
     peer->iface->conv = conv;
   }
@@ -925,6 +952,52 @@ static int ircd_charset_cb(INTERFACE *srv, struct peer_t *peer, const char *lcni
 }
 #endif
 
+BINDING_TYPE_ircd_client_cmd(ircd_help_cb);
+static int ircd_help_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnick,
+			const char *user, const char *host, const char *vhost,
+			modeflag eum, int argc, const char **argv)
+{ /* args: [<command>] */
+  CLIENT *cl = ((struct peer_priv *)peer->iface->data)->link->cl, *me;
+  userflag uf = 0;
+  struct binding_t *b;
+  char prefix[SHORT_STRING];
+
+  /* convert modeflag to userflag */
+  if (cl->umode & A_ADMIN)
+    uf = (U_HALFOP | U_OP | U_MASTER);
+  else if (cl->umode & A_OP)
+    uf = (U_HALFOP | U_OP);
+  else if (cl->umode & A_HALFOP)
+    uf = U_HALFOP;
+  if (cl->umode & A_REGISTERED)
+    uf |= U_FRIEND;
+  me = ircd_find_client(NULL, NULL);
+  /* return help */
+  if (argc > 0) {
+    snprintf(prefix, sizeof(prefix), ":%s 705 %s %s :", me->lcnick, peer->dname,
+	     argv[0]); /* RPL_HELPTXT */
+    b = Check_Bindtable (BTIrcdClientCmd, argv[0], uf, U_ANYCH, NULL);
+    if (b)
+      Get_Help_L("=ircd", b->key, peer->iface, U_NEGATE, uf, BTIrcdClientCmd,
+		 prefix, 1, -1, ""); //FIXME: LANG support
+    ircd_do_unumeric(cl, RPL_HELPTXT, cl, 0, argv[0]);
+    Get_Help_L("=ircd", b ? b->key : argv[0], peer->iface, U_NEGATE, uf,
+	       BTIrcdClientCmd, prefix, 1, 2, ""); //FIXME: LANG support
+    // RPL_HELPTXT empty line ?
+    return ircd_do_unumeric(cl, RPL_ENDOFHELP, cl, 0, argv[0]);
+  } else {
+    // RPL_HELPSTART ?
+    snprintf(prefix, sizeof(prefix), ":%s 705 %s index :", me->lcnick, peer->dname);
+    b = NULL;
+    while ((b = Check_Bindtable (BTIrcdClientCmd, NULL, uf, U_ANYCH, b)))
+    {
+      Get_Help_L("=ircd", b->key, peer->iface, U_NEGATE, uf, BTIrcdClientCmd,
+		 prefix, -1, 1, ""); //FIXME: LANG support
+    }
+    return ircd_do_unumeric(cl, RPL_ENDOFHELP, cl, 0, "index");
+  }
+}
+
 
 /* ---------------------------------------------------------------------------
  * Common external functions.
@@ -962,6 +1035,7 @@ void ircd_client_proto_end (void)
 #if IRCD_USES_ICONV
   Delete_Binding ("ircd-client-cmd", &ircd_charset_cb, NULL);
 #endif
+  Delete_Binding ("ircd-client-cmd", &ircd_help_cb, NULL);
 }
 
 void ircd_client_proto_start (void)
@@ -969,7 +1043,7 @@ void ircd_client_proto_start (void)
   Add_Binding ("ircd-check-message", "*", 0, 0, &ichmsg_ircd, NULL);
   Add_Binding ("ircd-client-cmd", "oper", 0, 0, &ircd_oper_cb, NULL);
   Add_Binding ("ircd-client-cmd", "quit", 0, 0, &ircd_quit_cb, NULL);
-  Add_Binding ("ircd-client-cmd", "squit", 0, 0, &ircd_squit_cb, NULL);
+  Add_Binding ("ircd-client-cmd", "squit", 0, U_HALFOP, &ircd_squit_cb, NULL);
   Add_Binding ("ircd-client-cmd", "part", 0, 0, &ircd_part_cb, NULL);
   Add_Binding ("ircd-client-cmd", "topic", 0, 0, &ircd_topic_cb, NULL);
   Add_Binding ("ircd-client-cmd", "invite", 0, 0, &ircd_invite_cb, NULL);
@@ -977,22 +1051,24 @@ void ircd_client_proto_start (void)
   Add_Binding ("ircd-client-cmd", "servlist", 0, 0, &ircd_servlist_cb, NULL);
   Add_Binding ("ircd-client-cmd", "who", 0, 0, &ircd_who_cb, NULL);
 #ifdef IRCD_ENABLE_KILL
-  Add_Binding ("ircd-client-cmd", "kill", 0, 0, &ircd_kill_cb, NULL);
+  Add_Binding ("ircd-client-cmd", "kill", 0, U_HALFOP, &ircd_kill_cb, NULL);
 #endif
   Add_Binding ("ircd-client-cmd", "away", 0, 0, &ircd_away_cb, NULL);
 #ifdef IRCD_ENABLE_REHASH
-  Add_Binding ("ircd-client-cmd", "rehash", 0, 0, &ircd_rehash_cb, NULL);
+  Add_Binding ("ircd-client-cmd", "rehash", 0, U_HALFOP, &ircd_rehash_cb, NULL);
 #endif
 #ifdef IRCD_ENABLE_DIE
-  Add_Binding ("ircd-client-cmd", "die", U_HALFOP, 0, &ircd_die_cb, NULL);
+  Add_Binding ("ircd-client-cmd", "die", U_HALFOP, U_HALFOP, &ircd_die_cb, NULL);
 #endif
 #ifdef IRCD_ENABLE_RESTART
-  Add_Binding ("ircd-client-cmd", "restart", U_HALFOP, 0, &ircd_restart_cb, NULL);
+  Add_Binding ("ircd-client-cmd", "restart", U_HALFOP, U_HALFOP, &ircd_restart_cb, NULL);
 #endif
   Add_Binding ("ircd-client-cmd", "userhost", 0, 0, &ircd_userhost_cb, NULL);
   Add_Binding ("ircd-client-cmd", "ison", 0, 0, &ircd_ison_cb, NULL);
 #if IRCD_USES_ICONV
   Add_Binding ("ircd-client-cmd", "charset", 0, 0, &ircd_charset_cb, NULL);
 #endif
+  BTIrcdClientCmd = Add_Bindtable ("ircd-client-cmd", B_UNIQ);
+  Add_Binding ("ircd-client-cmd", "help", 0, 0, &ircd_help_cb, NULL);
 }
 #endif
