@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2010-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -404,9 +404,14 @@ static void _ircd_broadcast_msglist_mark(IRCD *ircd, const char *nick,
 	  lnk->cl->via->p.iface->ift |= I_PENDING;
       need_unmark = 1;
       _ircd_mark_message_target(ircd->iface, nick, tlist[i]);
-    } else
+    } else {
       tcl->cs->via->p.iface->ift |= I_PENDING;
-      //FIXME: do alternate way for exact targets?
+#if IRCD_MULTICONNECT
+      /* still do alternate way for exact targets for better delivery chance */
+      if (tcl->cs->alt)
+	tcl->cs->alt->p.iface->ift |= I_PENDING;
+#endif
+    }
   }
   if (need_unmark)
     for (lnk = ircd->token[0]->c.lients; lnk; lnk = lnk->prev) /* no locals */
@@ -422,6 +427,8 @@ static int _ircd_broadcast_msglist_new (IRCD *ircd, struct peer_priv *via,
   register LINK *srv;
   int rc;
 
+  if (id < 0)
+    id = ircd_new_id(ircd->token[token]);
   for (srv = ircd->servers; srv; srv = srv->prev) /* preset to ignore later */
     if (!(srv->cl->umode & A_MULTI) || srv->cl->via == via ||
 	srv->cl->x.a.token == token)
@@ -444,7 +451,7 @@ static int _ircd_broadcast_msglist_new (IRCD *ircd, struct peer_priv *via,
 #endif
 
 static int _ircd_broadcast_msglist_old (IRCD *ircd, struct peer_priv *via,
-			unsigned short token, const char *nick, int all,
+			unsigned short token, const char *nick,
 			const char *targets, const char **tlist, size_t s,
 			const char *mode, const char *msg)
 {
@@ -454,7 +461,7 @@ static int _ircd_broadcast_msglist_old (IRCD *ircd, struct peer_priv *via,
   for (srv = ircd->servers; srv; srv = srv->prev) /* preset to ignore later */
     if (srv->cl->via == via ||
 #if IRCD_MULTICONNECT
-	(!all && (srv->cl->umode & A_MULTI)) ||
+	(srv->cl->umode & A_MULTI) ||
 #endif
 	srv->cl->x.a.token == token)
       srv->cl->via->p.iface->ift |= I_PENDING;
@@ -463,7 +470,7 @@ static int _ircd_broadcast_msglist_old (IRCD *ircd, struct peer_priv *via,
   for (srv = ircd->servers; srv; srv = srv->prev) /* reset them now */
     if (srv->cl->via == via ||
 #if IRCD_MULTICONNECT
-	(!all && (srv->cl->umode & A_MULTI)) ||
+	(srv->cl->umode & A_MULTI) ||
 #endif
 	srv->cl->x.a.token == token)
       srv->cl->via->p.iface->ift &= ~I_PENDING;
@@ -496,10 +503,10 @@ static inline CLIENT *_ircd_find_msg_target (const char *target,
     if (tgt && !CLIENT_IS_SERVER(tgt))
       tgt = NULL;
     if (tgt && !CLIENT_IS_ME(tgt))
-      return (tgt);
+      return (NULL);
     /* process user[%host] now */
     h = strchr(target, '%');
-    if (h > c)
+    if (h && h > c)
       h = NULL;
     tgt = ircd_find_by_userhost(NULL, 0, target, h ? h - target : c - target,
 				h ? h + 1 : NULL, h ? c - h - 1 : 0);
@@ -666,9 +673,9 @@ static int ircd_privmsg_cb(INTERFACE *srv, struct peer_t *peer, const char *lcni
   }
   if (s)
   {
-    _ircd_broadcast_msglist_new (ircd, NULL, 0, ircd_new_id(), peer->dname,
+    _ircd_broadcast_msglist_new (ircd, NULL, 0, ircd_new_id(NULL), peer->dname,
 				 targets, tlist, s, "PRIVMSG", argv[1]);
-    _ircd_broadcast_msglist_old (ircd, NULL, 0, peer->dname, 0,
+    _ircd_broadcast_msglist_old (ircd, NULL, 0, peer->dname,
 				 targets, tlist, s, "PRIVMSG", argv[1]);
   }
   return 1;
@@ -769,9 +776,9 @@ static int ircd_notice_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnic
   }
   if (s)
   {
-    _ircd_broadcast_msglist_new (ircd, NULL, 0, ircd_new_id(), peer->dname,
+    _ircd_broadcast_msglist_new (ircd, NULL, 0, ircd_new_id(NULL), peer->dname,
 				 targets, tlist, s, "NOTICE", argv[1]);
-    _ircd_broadcast_msglist_old (ircd, NULL, 0, peer->dname, 0,
+    _ircd_broadcast_msglist_old (ircd, NULL, 0, peer->dname,
 				 targets, tlist, s, "NOTICE", argv[1]);
   }
   return 1;
@@ -800,9 +807,9 @@ static int ircd_squery_cb(INTERFACE *srv, struct peer_t *peer, const char *lcnic
     return 1;
   }
 #endif
-  _ircd_broadcast_msglist_new(ircd, NULL, 0, ircd_new_id(),
-			      peer->dname, argv[0], argv, 1, "SQUERY", argv[1]);
-  _ircd_broadcast_msglist_old(ircd, NULL, 0, peer->dname, 0,
+  _ircd_broadcast_msglist_new(ircd, NULL, 0, ircd_new_id(NULL), peer->dname,
+			      argv[0], argv, 1, "SQUERY", argv[1]);
+  _ircd_broadcast_msglist_old(ircd, NULL, 0, peer->dname,
 			      argv[0], argv, 1, "SQUERY", argv[1]);
   return 1;
 }
@@ -834,13 +841,22 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
 	  argc);
     return ircd_recover_done(pp, "Invalid number of parameters");
   }
+#if IRCD_MULTICONNECT
+  if (pp->link->cl->umode & A_MULTI)
+  {
+    ERROR("ircd:illegal PRIVMSG command via %s", peer->dname);
+    return ircd_recover_done(pp, "illegal PRIVMSG command");
+  }
+#endif
   cl = _ircd_find_client_lc(ircd, lcsender);
   for (c = (char *)argv[0]; c; c = cnext)
   {
     if ((cnext = strchr(c, ',')))
       *cnext++ = 0;
     if (s == max_targets) {
-      _ircd_broadcast_msglist_old(ircd, pp, token, sender, 1,
+      _ircd_broadcast_msglist_new(ircd, pp, token, -1, sender,
+				  targets, tlist, s, "PRIVMSG", argv[1]);
+      _ircd_broadcast_msglist_old(ircd, pp, token, sender,
 				  targets, tlist, s, "PRIVMSG", argv[1]);
       s = s2 = 0;
     }
@@ -904,8 +920,12 @@ static int ircd_privmsg_sb(INTERFACE *srv, struct peer_t *peer, unsigned short t
     }
   }
   if (s)
-    _ircd_broadcast_msglist_old(ircd, pp, token, sender, 1,
+  {
+    _ircd_broadcast_msglist_new(ircd, pp, token, -1, sender,
 				targets, tlist, s, "PRIVMSG", argv[1]);
+    _ircd_broadcast_msglist_old(ircd, pp, token, sender,
+				targets, tlist, s, "PRIVMSG", argv[1]);
+  }
   return 1;
 }
 
@@ -931,13 +951,22 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 		peer->dname, argc);
     return (1);
   }
+#if IRCD_MULTICONNECT
+  if (pp->link->cl->umode & A_MULTI)
+  {
+    ERROR("ircd:illegal NOTICE command via %s", peer->dname);
+    return ircd_recover_done(pp, "illegal NOTICE command");
+  }
+#endif
   cl = _ircd_find_client_lc(ircd, lcsender);
   for (c = (char *)argv[0]; c; c = cnext)
   {
     if ((cnext = strchr(c, ',')))
       *cnext++ = 0;
     if (s == max_targets) {
-      _ircd_broadcast_msglist_old(ircd, pp, token, sender, 1,
+      _ircd_broadcast_msglist_new(ircd, pp, token, -1, sender,
+				  targets, tlist, s, "NOTICE", argv[1]);
+      _ircd_broadcast_msglist_old(ircd, pp, token, sender,
 				  targets, tlist, s, "NOTICE", argv[1]);
       s = s2 = 0;
     }
@@ -966,7 +995,10 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 	Add_Request(I_LOG, "*", F_WARN, "ircd:invalid NOTICE target %s via %s",
 		    c, peer->dname);
       } else if (!CLIENT_IS_REMOTE(tcl)) {
-	if (CLIENT_IS_SERVICE(cl))
+	if (CLIENT_IS_SERVER(cl))
+	  New_Request(tcl->via->p.iface, 0, ":%s NOTICE %s :%s", sender, c,
+		      argv[1]);
+	else if (CLIENT_IS_SERVICE(cl))
 	  New_Request(tcl->via->p.iface, 0, ":%s@%s NOTICE %s :%s", sender,
 		      cl->cs->lcnick, c, argv[1]);
 	else
@@ -997,8 +1029,12 @@ static int ircd_notice_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
     }
   }
   if (s)
-    _ircd_broadcast_msglist_old(ircd, pp, token, sender, 1,
+  {
+    _ircd_broadcast_msglist_new(ircd, pp, token, -1, sender,
 				targets, tlist, s, "NOTICE", argv[1]);
+    _ircd_broadcast_msglist_old(ircd, pp, token, sender,
+				targets, tlist, s, "NOTICE", argv[1]);
+  }
   return 1;
 }
 
@@ -1016,6 +1052,13 @@ static int ircd_squery_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
 	  argc);
     return ircd_recover_done(pp, "Invalid number of parameters");
   }
+#if IRCD_MULTICONNECT
+  if (pp->link->cl->umode & A_MULTI)
+  {
+    ERROR("ircd:illegal SQUERY command via %s", peer->dname);
+    return ircd_recover_done(pp, "illegal SQUERY command");
+  }
+#endif
   //cl = _ircd_find_client_lc((IRCD *)srv->data, lcsender);
   if (!(tcl = _ircd_find_msg_target(argv[0], pp)) ||
       !CLIENT_IS_SERVICE(tcl)) {
@@ -1027,8 +1070,12 @@ static int ircd_squery_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
     New_Request(tcl->via->p.iface, 0, ":%s SQUERY %s :%s", sender, argv[0], argv[1]);
   else
 #endif
-    _ircd_broadcast_msglist_old((IRCD *)srv->data, pp, token, sender, 1,
+  {
+    _ircd_broadcast_msglist_new((IRCD *)srv->data, pp, token, -1, sender,
+				argv[1], &argv[1], 1, "SQUERY", argv[2]);
+    _ircd_broadcast_msglist_old((IRCD *)srv->data, pp, token, sender,
 				argv[0], argv, 1, "SQUERY", argv[1]);
+  }
   return (1);
 }
 
@@ -1131,7 +1178,7 @@ static int ircd_iprivmsg(INTERFACE *srv, struct peer_t *peer, unsigned short tok
   if (s) {
     _ircd_broadcast_msglist_new(ircd, pp, token, id, sender,
 				targets, tlist, s, "PRIVMSG", argv[2]);
-    _ircd_broadcast_msglist_old(ircd, pp, token, sender, 0,
+    _ircd_broadcast_msglist_old(ircd, pp, token, sender,
 				targets, tlist, s, "PRIVMSG", argv[2]);
   }
   return 1;
@@ -1197,7 +1244,10 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
 	Add_Request(I_LOG, "*", F_WARN, "ircd:invalid INOTICE target %s via %s",
 		    c, peer->dname);
       } else if (!CLIENT_IS_REMOTE(tcl)) {
-	if (CLIENT_IS_SERVICE(cl))
+	if (CLIENT_IS_SERVER(cl))
+	  New_Request(tcl->via->p.iface, 0, ":%s NOTICE %s :%s", sender, c,
+		      argv[2]);
+	else if (CLIENT_IS_SERVICE(cl))
 	  New_Request(tcl->via->p.iface, 0, ":%s@%s NOTICE %s :%s", sender,
 		      cl->cs->lcnick, c, argv[2]);
 	else
@@ -1230,7 +1280,7 @@ static int ircd_inotice(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   if (s) {
     _ircd_broadcast_msglist_new(ircd, pp, token, id, sender,
 				targets, tlist, s, "NOTICE", argv[2]);
-    _ircd_broadcast_msglist_old(ircd, pp, token, sender, 0,
+    _ircd_broadcast_msglist_old(ircd, pp, token, sender,
 				targets, tlist, s, "NOTICE", argv[2]);
   }
   return 1;
@@ -1270,7 +1320,7 @@ static int ircd_isquery(INTERFACE *srv, struct peer_t *peer, unsigned short toke
   {
     _ircd_broadcast_msglist_new(ircd, pp, token, id, sender,
 				argv[1], &argv[1], 1, "SQUERY", argv[2]);
-    _ircd_broadcast_msglist_old(ircd, pp, token, sender, 0,
+    _ircd_broadcast_msglist_old(ircd, pp, token, sender,
 				argv[1], &argv[1], 1, "SQUERY", argv[2]);
   }
   return (1);
