@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003-2014  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2003-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -217,7 +217,7 @@ static ssize_t textlog_add_buf (logfile_t *log, char *text, size_t sz,
 	       log->path, str);
 #else
 	if (strerror_r(x, log->buf, sizeof(log->buf)) != 0)
-	  snprint(log->buf, sizeof(log->buf), "(failed to decode err=%d)", x);
+	  snprintf(log->buf, sizeof(log->buf), "(failed to decode err=%d)", x);
 	ERROR ("Couldn't sync logfile %s (%s), abort logging to it.",
 	       log->path, log->buf);
 #endif
@@ -684,7 +684,7 @@ static void do_rotate (logfile_t *log)
     ERROR ("Couldn't rotate %s: %s", log->path, str);
 #else
     if (strerror_r(x, path, sizeof(path)) != 0)
-      snprint(path, sizeof(path), "(failed to decode err=%d)", x);
+      snprintf(path, sizeof(path), "(failed to decode err=%d)", x);
     ERROR ("Couldn't rotate %s: %s", log->path, path);
 #endif
     return;
@@ -1037,7 +1037,10 @@ static void logrotate_reset (void)
     strcpy (logrotate_hr, "0");
   else if (x > 23)
     strcpy (logrotate_hr, "23");
-  x = atoi (logrotate_time+2);
+  if (logrotate_time[1])
+    x = atoi (logrotate_time+2);
+  else
+    x = 0;
   if (x > 59)
     x = 59;
   snprintf(logrotate_min, 3, "%d", x);
