@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2016-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -49,7 +49,7 @@
 struct sslbuff {
   BIO *bio;
   size_t bufptr, inbuf;
-  char buf[2*MB_LEN_MAX*MESSAGEMAX];
+  char buf[2*MBMESSAGEMAX];
 };
 
 struct connchain_buffer {
@@ -108,7 +108,7 @@ static void _s_check_saved_buffer(idx_t id, struct connchain_buffer *buf)
   o = 0;
   o = Connchain_Put(&buf->saved_chain, id, NULL, &o);
   /* kill chain when done */
-  if (i < 0 && o < 0 && Connchain_Get(&buf->saved_chain, id, NULL, 0))
+  if (i < 0 && o < 0 && Connchain_Get(&buf->saved_chain, -1, NULL, 0))
     buf->saved_chain = NULL;
   DBG("ssl: cleared old chain");
 }
@@ -133,7 +133,9 @@ static ssize_t _ssl_try_send_buffers(struct connchain_i **ch, idx_t id,
   /* DBG("ssl: tried to send data, size=%zu sent=%zd", so, i); */
   if (i < 0)
     return (i);
-  if (i > 0 && buf->out.inbuf > 0)
+  if (buf->out.inbuf == buf->out.bufptr) /* Connchain_Put() returns 1 in this case */
+    i = 0;
+  if (i > 0)
     dprint(6, "ssl: sent encrypted data, size=%zd", i);
   if (so == 0)			/* done */
     buf->out.bufptr = buf->out.inbuf = 0;
@@ -220,6 +222,7 @@ static ssize_t _ccfilter_S_send(struct connchain_i **ch, idx_t id, const char *s
       return (0);
     return Connchain_Put (ch, id, str, sz); /* ask next link to flush then */
   }
+retry:
   i = _ssl_try_send_buffers(ch, id, buf, TRUE); /* try to push buffers now */
   if (i < 0)
     return (i);
@@ -234,7 +237,11 @@ static ssize_t _ccfilter_S_send(struct connchain_i **ch, idx_t id, const char *s
     if (!_ssl_check_input_from_chain(ch, id, b)) /* SSL may wait for data */
       return Connchain_Put (ch, id, str, sz); /* if not SSL then bypass data */
     DBG("ssl: handshake is in progress");
-    if (!SSL_is_init_finished(buf->ssl))
+    if ((*b)->error < 0)
+      return ((*b)->error);	/* there might be an error from connchain */
+    if (i > 0)
+      goto retry;		/* we have send something to socket so retry */
+    if (!buf->check_done)
       return (0);
   }
   /* ready to process input data */
@@ -247,11 +254,10 @@ static ssize_t _ccfilter_S_send(struct connchain_i **ch, idx_t id, const char *s
     i = 0;
   }
   so = i;
-  if (buf->out.inbuf != 0) {	/* trying to push buffer again */
-    i = _ssl_try_send_buffers(ch, id, buf, TRUE);
-    if (i < 0)
-      return (i);
-  }
+  /* trying to push buffer again */
+  i = _ssl_try_send_buffers(ch, id, buf, TRUE);
+  if (i < 0)
+    return (i);
   return (so);
 }
 
@@ -283,6 +289,7 @@ static ssize_t _ccfilter_S_recv(struct connchain_i **ch, idx_t id, char *str,
       buf->in.bufptr += i;
     return (i);
   }
+retry:
   if (!_ssl_check_input_from_chain(ch, id, b)) { /* not a SSL data */
     i = (buf->in.inbuf - buf->in.bufptr);
     if (i > (ssize_t)sz)
@@ -299,6 +306,8 @@ static ssize_t _ccfilter_S_recv(struct connchain_i **ch, idx_t id, char *str,
     i = _ssl_try_send_buffers(ch, id, buf, TRUE); /* SSL may have data to send */
     if (i < 0)
       return (i);
+    if (i > 0)
+      goto retry;		/* we have send something to socket so retry */
   }
   i = SSL_read(buf->ssl, str, sz);
   if (i <= 0) {			/* some error in SSL_read */
@@ -310,7 +319,7 @@ static ssize_t _ccfilter_S_recv(struct connchain_i **ch, idx_t id, char *str,
     return (i);
   ERROR("ssl: got %zd from connection chain, terminating", i);
 finish_filter:
-  if (buf->saved_chain != NULL && Connchain_Get(&buf->saved_chain, id, NULL, 0))
+  if (buf->saved_chain != NULL && Connchain_Get(&buf->saved_chain, -1, NULL, 0))
     buf->saved_chain = NULL;
   i = (*b)->error;
   if (i == 0)
@@ -435,7 +444,11 @@ static iftype_t module_signal (INTERFACE *iface, ifsig_t sig)
     SSL_CTX_free(ctx);
     ctx = NULL;
     /* deinit library */
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     ERR_remove_state(0);
+#else
+    FIPS_mode_set(0);
+#endif
     ENGINE_cleanup();
     CONF_modules_unload(1);
     ERR_free_strings();
@@ -510,7 +523,11 @@ SigFunction ModuleInit (char *args)
 {
   CheckVersion;
   /* lib init */
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
   SSL_library_init();
+#else
+  OPENSSL_init_ssl(0, NULL);
+#endif
   SSL_load_error_strings();
   ERR_load_BIO_strings();
   OpenSSL_add_all_algorithms();
